from tqdm import trange

assert (3 / 2) * 2 == 3, "Use Python 3"


def main():
    with open('./flag.iso.enc', 'rb') as f:
        enc_data = f.read()

    assert len(enc_data) == 1052672  # sanity check

    N = 40000

    print('[+] Make sure you are running this with pypy3, else you will be sad...')
    try:
        from recovered_poly import poly, poly_size, BACKWARD_DEPTH
        print(f'[+] Using cached poly recovered from last {BACKWARD_DEPTH} bytes')
    except ImportError:
        BACKWARD_DEPTH = 100000
        assert BACKWARD_DEPTH >= 2 * N
        print(f'[+] Starting Berlekamp-Massey on the last {BACKWARD_DEPTH} bytes; once per bit')
        print(f"    ... brew a cup of coffee and come back. It's gonna take a few minutes")
        poly = []
        poly_size = []
        for bit_position in trange(8):
            p, ps = berlekamp_massey_algorithm(
                [(x >> bit_position) & 1
                 for x in enc_data[-BACKWARD_DEPTH:]])
            poly.append(p)
            poly_size.append(ps)
        with open('recovered_poly.py', 'w') as f:
            f.write(f'poly = {poly}\n')
            f.write(f'poly_size = {poly_size}\n')
            f.write(f'BACKWARD_DEPTH = {BACKWARD_DEPTH}\n')

    print(f'[+] Have a poly of size {poly_size}')
    assert all(p <= N for p in poly_size), (
        f"Somewhere in the last {BACKWARD_DEPTH} bytes, a non-zero existed")

    assert all(p == sorted(p, reverse=True) for p in poly)
    assert all(p[-1] == 0 for p in poly)

    print('[+] Reversing the LFSR to recover key material')
    print("    ... brew yet another cup of coffee and come back. It's gonna take a few minutes")
    keys = []
    for bit_position in trange(8):
        rev_key = [(x >> bit_position) & 1 for x in enc_data[-N:][::-1]]
        for _ in trange(len(enc_data) - len(rev_key)):
            o = 0
            for p in poly[bit_position][:-1]:
                o ^= rev_key[-p]
            rev_key.append(o)
        assert len(rev_key) == len(enc_data)
        key = rev_key[::-1]
        assert key[-N:] == [(x >> bit_position) & 1 for x in enc_data[-N:]]
        assert key[-BACKWARD_DEPTH:] == [(x >> bit_position) & 1
                                         for x in enc_data[-BACKWARD_DEPTH:]]
        keys.append(key)

    print('[+] Reversed key material.')
    key = [
        sum((1 << bs) * keys[bs][i] for bs in range(8))
        for i in range(len(enc_data))
    ]
    print('[+] Merged bits.')

    plaintext = bytes(k ^ e for k, e in zip(key, enc_data))
    print('[+] Recovered plaintext.')

    with open('output.iso', 'wb') as f:
        f.write(plaintext)
    print('[+] Saved to output.iso!')


def berlekamp_massey_algorithm(sequence):
    # This function is modified from
    # https://raw.githubusercontent.com/bozhu/BMA/master/bma.py
    #
    # Original copyright notice that allows copying and modification:
    """Copyright (C) 2012 Bo Zhu http://about.bozhu.me

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE."""
    N = len(sequence)
    s = sequence[:]

    for k in range(N):
        if s[k] == 1:
            break
    f = set([k + 1, 0])  # use a set to denote polynomial
    l = k + 1

    g = set([0])
    a = k
    b = 0

    for n in range(k + 1, N):
        d = 0
        for ele in f:
            d ^= s[ele + n - l]

        if d == 0:
            b += 1
        else:
            if 2 * l > n:
                f ^= set([a - b + ele for ele in g])
                b += 1
            else:
                temp = f.copy()
                f = set([b - a + ele for ele in f]) ^ g
                l = n + 1 - l
                g = temp
                a = b
                b = n - l + 1

    return (sorted(f, reverse=True), l)


if __name__ == '__main__':
    main()
